@@ -142,8 +142,10 @@ decode_bytes_with_escapes(Parser *p, const char *s, Py_ssize_t len, Token *t)
    If the string is an f-string, set *fstr and *fstrlen to the unparsed
    string object.  Return 0 if no errors occurred.  */
 int
-_Mys_PyPegen_parsestr(Parser *p, int *bytesmode, int *rawmode, PyObject **result,
-                  const char **fstr, Py_ssize_t *fstrlen, Token *t)
+_Mys_PyPegen_parsestr(Parser *p, int *bytesmode, int *rawmode, int *remode, int *cmode,
+                      int *is_char,
+                      PyObject **result, const char **fstr, Py_ssize_t *fstrlen,
+                      PyObject **reflags, Token *t)
 {
     const char *s = PyBytes_AsString(t->bytes);
     if (s == NULL) {
@@ -155,10 +157,15 @@ _Mys_PyPegen_parsestr(Parser *p, int *bytesmode, int *rawmode, PyObject **result
     int fmode = 0;
     *bytesmode = 0;
     *rawmode = 0;
+    *remode = 0;
+    *cmode = 0;
+    *is_char = 0;
     *result = NULL;
     *fstr = NULL;
+    *reflags = NULL;
+
     if (Py_ISALPHA(quote)) {
-        while (!*bytesmode || !*rawmode) {
+        while (!*bytesmode || !*rawmode || !*remode || !*cmode) {
             if (quote == 'b' || quote == 'B') {
                 quote =(unsigned char)*++s;
                 *bytesmode = 1;
@@ -169,17 +176,38 @@ _Mys_PyPegen_parsestr(Parser *p, int *bytesmode, int *rawmode, PyObject **result
             else if (quote == 'r' || quote == 'R') {
                 quote = (unsigned char)*++s;
                 *rawmode = 1;
+
+                if (quote == 'e' || quote == 'E') {
+                    quote = (unsigned char)*++s;
+                    *remode = 1;
+                }
             }
             else if (quote == 'f' || quote == 'F') {
                 quote = (unsigned char)*++s;
                 fmode = 1;
             }
+            else if (quote == 'c' || quote == 'C') {
+                quote = (unsigned char)*++s;
+                *rawmode = 1;
+                *cmode = 1;
+            }
             else {
                 break;
             }
         }
     }
 
+
+    if (quote == '\'') {
+        if (fmode || *bytesmode || *rawmode || *remode || *cmode) {
+            p->error_indicator = 1;
+            RAISE_SYNTAX_ERROR("characters cannot have a prefix");
+            return -1;
+        }
+
+        *is_char = 1;
+    }
+
     /* fstrings are only allowed in Python 3.6 and greater */
     if (fmode && p->feature_version < 6) {
         p->error_indicator = 1;
@@ -202,12 +230,26 @@ _Mys_PyPegen_parsestr(Parser *p, int *bytesmode, int *rawmode, PyObject **result
         PyErr_SetString(PyExc_OverflowError, "string to parse is too long");
         return -1;
     }
+    if (*remode) {
+        size_t flagslen = 0;
+        while (s[len - flagslen - 1] != quote) {
+            flagslen++;
+        }
+        len -= flagslen;
+        *reflags = PyUnicode_DecodeUTF8Stateful(&s[len], flagslen, NULL, NULL);
+    }
     if (s[--len] != quote) {
         /* Last quote char must match the first. */
         PyErr_BadInternalCall();
         return -1;
     }
     if (len >= 4 && s[0] == quote && s[1] == quote) {
+        if (*is_char) {
+            p->error_indicator = 1;
+            RAISE_SYNTAX_ERROR("characters cannot be triple quoted");
+            return -1;
+        }
+        
         /* A triple quoted string. We've already skipped one quote at
            the start and one at the end of the string. Now skip the
            two at the start. */
@@ -257,6 +299,7 @@ _Mys_PyPegen_parsestr(Parser *p, int *bytesmode, int *rawmode, PyObject **result
             *result = decode_unicode_with_escapes(p, s, len, t);
         }
     }
+
     return *result == NULL ? -1 : 0;
 }
 
